#include <cstdio>
#include <string>
#include <list>
#include <set>
#include <map>

#include "autoptr"
#include "riffhead.hh"

using namespace std;

#define bufsize 512

#include "stat.h"



bool quiet = false;
static bool filepos=false;

static set<string> listignores;

struct tieto
{
    int ind;
    string field, info;

    SizeType startpos;
    SizeType bytes, overuse, excess;
    
    const string getstring() const
    {
        const string &k = field;
        
        char BufO[64]=""; if(overuse)sprintf(BufO, ", "SizeFormat" OVERUSE", overuse);
        char BufE[64]=""; if(excess)sprintf(BufE, ", "SizeFormat" data", excess);
        
        char Buf[512];
        string result(Buf, sprintf(Buf, "%*s", ind,""));
        if(filepos)
            result += string(Buf, sprintf(Buf, SizeFormat": ", startpos));
        result += k;
        result += string(Buf,
          sprintf(Buf, " ("SizeFormat" bytes%s%s)",
            bytes,
            BufO,
            BufE));
        if(info.size()) { result += " ("; result += info; result += ')'; }
        return result;
    }
};

static list<tieto> tiedot;

unsigned ReadDW(FILE *fp)
{
    char Buf[4];
    fread(Buf, 1, 4, fp);
    return ((unsigned)((unsigned char)Buf[0]))
         | ((unsigned)((unsigned char)Buf[1]) << 8)
         | ((unsigned)((unsigned char)Buf[2]) << 16)
         | ((unsigned)((unsigned char)Buf[3]) << 24);
}
unsigned ReadW(FILE *fp)
{
    char Buf[2];
    fread(Buf, 1, 2, fp);
    return ((unsigned)((unsigned char)Buf[0]))
         | ((unsigned)((unsigned char)Buf[1]) << 8);
}
void PatchW(FILE *fp, SizeType pos, unsigned value)
{
    unsigned char Buf[2];
    Buf[0] = value & 255;
    Buf[1] = (value >> 8) & 255;
    SizeType oldpos = FTellFunc(fp);
    FSeekFunc(fp, pos, SEEK_SET);
    if(fwrite(Buf, 1, 2, fp) == 2)
        fprintf(stderr, "Patched ["SizeFormat"] with 0x%04X\n", pos, value);
    else
        fprintf(stderr, "Patch ["SizeFormat"] with 0x%04X FAILED\n", pos, value);
    FSeekFunc(fp, oldpos, SEEK_SET);
}
void PatchDW(FILE *fp, SizeType pos, unsigned value)
{
    unsigned char Buf[4];
    Buf[0] = value & 255;
    Buf[1] = (value >> 8) & 255;
    Buf[2] = (value >> 16) & 255;
    Buf[3] = (value >> 24) & 255;
    SizeType oldpos = FTellFunc(fp);
    FSeekFunc(fp, pos, SEEK_SET);
    if(fwrite(Buf, 1, 4, fp) == 4)
        fprintf(stderr, "Patched ["SizeFormat"] with 0x%08X\n", pos, value);
    else
        fprintf(stderr, "Patch ["SizeFormat"] with 0x%08X FAILED\n", pos, value);
    FSeekFunc(fp, oldpos, SEEK_SET);
}

class readermap: public map<string, autoptr<RiffHead> >, public ptrable {};
static readermap *readers = NULL;
static class poo{public:~poo(){delete readers;}} poo;

RiffHead::~RiffHead() {}
void RegisterFormat(const char *code, autoptr<RiffHead> instance)
{
    if(!readers)readers = new readermap;
    (*readers)[code] = instance;
}

void FindSomething(FILE *fp)
{
    static int ind=-1;struct ipp{int i;ipp():i(ind){ind+=1;}~ipp(){ind=i;}}ipp;
    
    if(ferror(fp))
    {
        perror("");
        return;
    }
    
    tieto data;
    data.startpos = FTellFunc(fp);
    data.ind = ind;
    
    char Buf[5] = "????";
    fread(Buf, 1, 4, fp);
    if(feof(fp))return;

    data.bytes = ReadDW(fp);

    data.field = Buf;

    SizeType pos = FTellFunc(fp);
    SizeType endpos = pos + data.bytes;
    
    unsigned skip = (data.bytes + 1) & (~1UL);
    pos += skip;
    if(pos & 1) ++pos;
    
    list<tieto>::iterator between = tiedot.end(); --between;
    
    
    if(data.field == "LIST"
    || data.field == "RIFF")
    {
        char Buf2[5];
        fread(Buf2, 1, 4, fp);
        Buf2[4] = 0;
        string s2 = Buf2;
        
        data.field += '(';
        data.field += Buf2;
        data.field += ')';
        
        if(listignores.find(s2) == listignores.end())
            while(!feof(fp) && !ferror(fp) && FTellFunc(fp) < endpos)
            {
                if(!quiet)
                {
                    double pos = FTellFunc(fp) * 100.0 / endpos;
                    sprintf(Buf2, "%4.1f", pos);
                    static string prosk;
                    if(prosk!=Buf2) {prosk=Buf2;fprintf(stderr,"%s%%\r",Buf2);fflush(stderr);}
                }
                FindSomething(fp);
            }
    }
    else
    {
        map<string, autoptr<RiffHead> >::iterator i = readers->find(data.field);
        if(i != readers->end())
        {
            data.info = i->second->Read(fp, endpos-FTellFunc(fp));
        }
    }
    
    SizeType pos2 = FTellFunc(fp);
    data.overuse = pos2 > endpos ? pos2-endpos : 0;
    data.excess  = pos2 < endpos ? endpos-pos2 : 0;
    
    if(data.overuse > 0)
        FSeekFunc(fp, pos, SEEK_SET);
    else if(data.excess > 0) // bufsize-8)
        FSeekFunc(fp, pos, SEEK_SET);
    /*else if(excess > 0)
    {
        char Buf[bufsize];
        fread(Buf, 1, excess, fp);
    }*/

    tiedot.insert(++between, data);
}

#include <unistd.h>
#include <cstdlib>

static void Handle(FILE *fp)
{
    if(!readers)readers = new readermap;
    
    {tieto tmp; tiedot.push_back(tmp);}
    
    {static char Buf[bufsize];setbuffer(fp,Buf,bufsize);}
    //setbuffer(fp, NULL, 64);
    
    while(!feof(fp))
        FindSomething(fp);
    
    tiedot.erase(tiedot.begin());
    
    list<tieto>::const_iterator i;
    for(i=tiedot.begin(); i!=tiedot.end(); ++i)
    {
        const string &s = i->getstring();
        fwrite(s.data(), 1, s.size(), stdout);
        putchar('\n');
    }
}

class dblregfieldmap: public map<string, string>, public ptrable {};
static dblregfieldmap *double_regfields = NULL;
static class dblpoo{public:~dblpoo(){delete double_regfields;}} dblpoo;
static map<string, double> double_patches;
void RegisterFieldDouble(const char *name, const char *description)
{
    if(!double_regfields) double_regfields = new dblregfieldmap;
    if(double_regfields->find(name) !=
       double_regfields->end())
    {
        fprintf(stderr, "Internal error: Attempt to register '%s' twice\n",
            name);
    }
    (*double_regfields)[name] = description;
}
bool IsDefined(const string& name)
{
    return double_patches.find(name) != double_patches.end();
}
double GetFieldDouble(const string& name)
{
    return double_patches[name];
}
static bool IsPatching()
{
    return !double_patches.empty();
}
static void DefineField(const string& name, const string& value)
{
    if(IsDefined(name))
    {
        fprintf(stderr, "Warning: Redefinition of '%s' as '%s'\n",
            name.c_str(), value.c_str());
    }
    if(double_regfields->find(name)
    != double_regfields->end())
    {
        double_patches[name] = strtod(value.c_str(), NULL);
        return;
    }
    fprintf(stderr, "Error: Attempt to define unknown field '%s'\n", name.c_str());
}
static void ListFieldDetails()
{
    for(map<string, string>::const_iterator 
        i = double_regfields->begin();
        i != double_regfields->end();
        ++i)
    {
        printf("    %s (%s) example: -m%s=%g\n",
            i->first.c_str(),
            i->second.c_str(),
            i->first.c_str(), 10.1
              );
    }
}

#include <getopt.h>

int main(int argc, char**argv)
{
    std::list<std::string> files;
    
    std::set<std::string> dump;
    
    for(;;)
    {
        int option_index = 0;
        static struct option long_options[] =
        {
            {"help",       0,0,'h'},
            {"version",    0,0,'V'},
            {"quiet",      0,0,'q'},
            {"listignore", 1,0,'n'},
            {"filepos",    0,0,'p'},
            {"modify",     1,0,'m'},
            {"dump",       1,0,'D'},
            {0,0,0,0}
        };
        
        int c = getopt_long(argc,argv, "hVqn:pm:D:", long_options, &option_index);
        if(c==-1) break;
        switch(c)
        {
            case 'V': printf("%s\n", VERSION); return 0;
            case 'h':
            {
                printf(
                    "This is riffdump-"VERSION"\n"
                    "Copyright (C) 1992,2006 Bisqwit (http://iki.fi/bisqwit/)\n"
                    "\n"
                    "Displays the structure of a RIFF file (AVI, WAV, ...)\n"
                    "\n"
                    "Usage: riffdump [<option> [<...>]] <file>\n"
                    "\n"
                    "Options:\n");
                
                printf(
                    "  -h, --help               This help.\n"
                    "  -V, --version            Displays version information.\n"
                    "  -q, --quiet              Disable displaying %% of completion.\n"
                    "  -n, --listignore <type>  Ignore LIST chunks of specified type,\n"
                    "                           Example: -nmovi\n"
                    "  -p, --filepos            Display filepos\n"
                    "  -m, --modify <field=value>\n"
                    "                           Modify AVI header\n"
                    "  -D, --dump <num>[,<...>] Dump streams\n"
                      );
                
                printf("\n"
                    "Fields understood by --modify:\n");
                ListFieldDetails();
                
                printf("\n"
                       "riffdump is a partner program of animmaker.\n"
                       "No warranty whatsoever.\n");
                return 0;
            }
            case 'n':
                listignores.insert(optarg);
                break;
            case 'D':
            {
                std::string rest = optarg;
                for(;;)
                {
                    std::string::size_type spacepos = rest.find(',');
                    if(spacepos == rest.npos)break;
                    dump.insert(rest.substr(0, spacepos));
                    if(spacepos+1 >= rest.size()) { rest = ""; break; }
                    rest = rest.substr(spacepos+1);
                }
                if(rest.size()) { dump.insert(rest); rest = ""; }
                break;
            }
            case 'm':
            {
                std::string value = optarg;
                std::string::size_type eqpos = value.find('=');
                if(eqpos == value.npos)
                {
                    fprintf(stderr, "riffdump: --modify: syntax error. Correct syntax example: -mfps=20\n");
                    return -1;
                }
                DefineField(value.substr(0, eqpos), value.substr(eqpos+1));
                break;
            }
            case 'q':
                quiet = true;
                break;
            case 'p':
                filepos = true;
                break;
        }
    }
    
    while(optind < argc)
        files.push_back(argv[optind++]);

    int errs = 0;
    
    if(files.empty())
    {
        if(isatty(STDIN_FILENO))
        {
            fprintf(stderr, "riffdump: AVI data not written to a terminal.\n");
            ++errs;
        }
        else
            Handle(stdin);
    }
    else
    {
        list<string>::const_iterator i;
        for(i=files.begin(); i!=files.end(); ++i)
        {
            if(*i == "-")
            {
                if(isatty(STDIN_FILENO))
                {
                    fprintf(stderr, "riffdump: AVI data not written to a terminal.\n");
                    ++errs;
                }
                else
                    Handle(stdin);
                continue;
            }
            const char *openmode = IsPatching() ? "rb+" : "rb";            
            FILE *fp = FOpenFunc(i->c_str(), openmode);
            if(!fp)
                { perror(i->c_str()); ++errs; continue; }
            Handle(fp);
            fclose(fp);
        }
    }
    return errs;
}
